open Syntax
open Util


let expand_nonce (s:string) (repl:int) : string = 
  if repl = 0 then 
    s
  else
    String.sub s 0 (String.length s - 1) ^ "#" ^ string_of_int repl

let expand_var (s:string) (repl:int) : string = 
  if repl = 0 then 
    s
  else
    s ^ "#" ^ string_of_int repl

let rec expand_repl_term (t:term) (repl:int): term = 
  match t with 
  | VNonce (e, n) -> 
      (match n with 
      | Nonce n -> t
      | NonceStar n -> VNonce (e, Nonce (expand_nonce n repl)))
  | VVar (e, x) -> VVar (e, expand_var x repl) 
  | VConst _ -> t
  | VKey _ -> t 
  | VPKey _ -> t
  | VVerKey _ -> t
  | VPair (e, u, v) -> VPair (e, expand_repl_term u repl, expand_repl_term v repl)
  | VInl (e, v) -> VInl (e, expand_repl_term v repl)
  | VInr (e, v) -> VInr (e, expand_repl_term v repl)
  | VHash (e, v) -> VHash (e, expand_repl_term v repl)
  | VSEnc (e, v, k) -> VSEnc (e, expand_repl_term v repl, k)
  | VAEnc (e, v, k) -> VAEnc (e, expand_repl_term v repl, k)
  | VSign (e, v, k) -> VSign (e, expand_repl_term v repl, k)

let expand_repl_constr (c:constr) (repl:int) : constr = 
  let u,v = c in 
  expand_repl_term u repl, expand_repl_term v repl



let expand_repl_env (g:var_env) (repl:int) : var_env = 
  VarEnv.fold (fun x t g' -> 
        let x' =  expand_var x repl in
        extend_var_env g' x' t)
  g (var_env_empty ())


let expand_repl_nonce_env (e:nonce_env) : nonce_env = 
  fun (x:nonce) -> 
    match x with 
    | Nonce x' ->
      if String.length x' >= 3 && (String.sub x' (String.length x' - 2) 2 = ("#" ^ string_of_int 1) || String.sub x' (String.length x' - 2) 2 = ("#" ^ string_of_int 2)) then
        e (NonceStar (String.sub x' 0 (String.length x' - 2) ^ "*"))
      else
        e x
    | NonceStar _  -> failwith "Should not look up Star variables at this point"

let expand_repl_constr_set_env (cse:constr_set_env) (repl:int) : constr_set_env = 
  let cs, g = cse in 
  let cs' = List.map (fun c -> expand_repl_constr c repl) cs in  
  (* let g' = expand_repl_env_types g in *)
  let g' = expand_repl_env g repl in 
  cs', g'

let expand_repl_constr_set_set (css:constr_set_set) (repl:int): constr_set_set = 
  print_verbose 2 "... Replicating Constraints";
  List.map (fun cse -> expand_repl_constr_set_env cse repl) css


exception Constraint_Error of string

(* Step 0 from algorithm *)
(* Check if initial refinements are of the correct form *)
let step0 (e:nonce_env) (e':nonce_env) (g:var_env) (d:key_env) (c:constr) : bool =  
  match c with 
  | VSEnc (_, _, k), VSEnc (_,_,k') -> 
      (match d k, d  k' with 
      | Some (TKey (LS,_)), Some (TKey (LS, _)) -> true
      | _ -> raise (Constraint_Error ("Both encryption keys must be S! " ^ constr2string c)))
  | VAEnc (_, t1, PKey k), VAEnc (_,t2,PKey k') -> 
      (match d k, d  k' with 
      | Some (TKey (LS,_)), Some (TKey (LS, _)) -> true
      | _ -> raise (Constraint_Error ("Both encryption keys must be S! " ^ constr2string c)))
  | VHash (_,t1), VHash (_, t2) -> true
  | VSign (_, t1, k), VSign (_,t2,k') -> 
      (match d k, d  k' with 
      | Some (TKey (LS,_)), Some (TKey (LS, _)) -> true
      | _ -> raise (Constraint_Error ("Both encryption keys must be S! " ^ constr2string c)))
  | _ -> raise ( Constraint_Error ("Illegal constraint   " ^ constr2string c))

(* Step 1 from algorithm *)
(* Initialize all variables with non-star refinement types *)
let step1 (g:var_env) (c:constr) : constr =
  let u,v = c in 
  let u' = instantiate g u true false in 
  let v' = instantiate g v false false in 
  u', v'

(* Step 2 from algorithm *)
(* unfold all pairs, public encryptions and signatures *)
let step2' (d:key_env) (c : constr) : constr_set * constr_set = 
  match c with 
  | VPair (_, t1, t2), VPair (_, t1', t2') -> [(t1, t1'); (t2,t2')], []
  | VSEnc (_, t1, k), VSEnc (_, t1', k') ->  
      if (k = k') then
        match d k with 
        | Some (TKey (LL,_ )) -> [(t1,t1')], []
        | _ -> [], [c]
      else
        failwith "Illegal state : encryption with different keys in constr"
  | VAEnc (_, t1, k), VAEnc (_, t1', k') ->  
      if (k = k') then
        let PKey k = k in 
        match d k with 
        | Some (TKey (LL, _)) -> [(t1,t1')], []
        | _ -> [], [c]
      else
        failwith "Illegal state : encryption with different keys in constr"
  | VSign (_, t1, k), VSign (_, t1', k') ->  
      if (k = k') then
        match d k with 
        | Some (TKey (LL, _))-> [(t1,t1')], []
        | Some (TKey (LS, _)) -> [(t1,t1')], [c]
        | _ -> failwith "Signature with H key. What happens here?"
      else
        failwith "Illegal state : encryption with different keys in constr"
  | VInl (_,t), VInl (_,t') -> [(t,t')], []
  | VInr (_,t), VInr (_,t') -> [(t,t')], []
  | _ -> [], [c]

(* Step 2 from algorithm *)
let rec step2 (d:key_env) (cs1:constr_set) (cs2:constr_set) : constr_set  = 
  let cs1', cs2' = List.fold_left (fun (cs1l, cs2l) c -> let c1, c2 = step2' d c in c1 @ cs1l, c2 @ cs2l) ([], cs2) cs1 in 
  if cs1' = [] then cs2'
  else step2 d cs1' cs2'


(* Step 3 from algorithm *)
let step3 (e:nonce_env) (e':nonce_env) (g:var_env) (d:key_env) (c:constr) : bool =  
  match c with 
  | VNonce (_, n), VNonce (_, m) ->
      (match e n, e' m  with 
       | Some LL , Some LL  -> true
       | _ -> raise (Constraint_Error ("Both nonces must be Low! " ^ constr2string c)))
  | VKey (_, k), VKey (_, k') ->
      (match d k, d  k' with 
      | Some (TKey (LL,_)), Some (TKey (LL, _)) -> true
      | _ -> raise (Constraint_Error ("Both keys must be Low! " ^ constr2string c)))
  | VPKey _, VPKey _ -> true
  | VVerKey _, VVerKey _ -> true
  | VConst _ , VConst _ -> true
  | VSEnc (_, _, k), VSEnc (_,_,k') -> 
      (match d k, d  k' with 
      | Some (TKey (LS,_)), Some (TKey (LS, _)) -> true
      | _ -> raise (Constraint_Error ("Both encryption keys must be S! " ^ constr2string c)))
  | VAEnc (_, t1, PKey k), VAEnc (_,t2,PKey k') -> 
      (match d k, d  k' with 
      | Some (TKey (LS,_)), Some (TKey (LS, _)) -> 
          if s_under_pair e d t1  && s_under_pair e' d t2 then
            true
          else
            raise (Constraint_Error ("No S value under pair!  " ^ constr2string c))
      | _ -> raise (Constraint_Error ("Both encryption keys must be S! " ^ constr2string c)))
  | VHash (_,t1), VHash (_, t2) ->
        if s_under_pair e d t1  && s_under_pair e' d t2 then
          true
        else
          raise (Constraint_Error ("No S value under pair!  " ^ constr2string c))
  | VSign (_, t1, k), VSign (_,t2,k') -> 
      (match d k, d  k' with 
      | Some (TKey (LS,_)), Some (TKey (LS, _)) -> true
      | _ -> raise (Constraint_Error ("Both encryption keys must be S! " ^ constr2string c)))
  | _ -> raise ( Constraint_Error ("Illegal constraint   " ^ constr2string c))



(* not very efficient *)
let occurs (x:var) (u:term) : bool =
  List.mem x (snd (collectnv_t u))


exception NoUnifier;;

let mgu (ul : (term*term) list) : (term*term) list =
  let rec mgu_aux (ul : (term*term) list) (vl : (term*term) list) : (term*term) list =
    match vl with
    | [] -> ul
    (*trivial case*)
    | ((u,v) :: l) when same_term u v -> mgu_aux ul l 
    (* occur check to prevent cycles *)
    | ((VVar (_,x),u) :: l) when (occurs x u) -> raise NoUnifier 
    (* replace *)
    | ((VVar (e,x),u) :: l) ->
       let ull = List.map (fun (y,w) -> (y,substitute [x,u] w)) ul in
       let vll = List.map
                   (fun (v,w) -> (substitute [x,u] v, substitute [x,u] w))
                   l
       in
       mgu_aux ((VVar (e,x),u) :: ull) vll
    (* transposition *)
    | ((u,VVar (e,x)) :: l) -> mgu_aux ul ((VVar (e,x),u)::l)
    (* decomposition/incompatible head symbols *)
    | ((VNonce (_, m), VNonce (_, n)) :: l) ->
       if m = n then
         mgu_aux ul l
       else
         raise NoUnifier
    | ((VConst (_, cst), VConst (_, cst')) :: l) ->
       if cst = cst' then
         mgu_aux ul l
       else
         raise NoUnifier
    | ((VKey (_, k), VKey (_, k')) :: l) ->
       if k = k' then
         mgu_aux ul l
       else
         raise NoUnifier
    | ((VPKey (_, pk), VPKey (_, pk')) :: l) ->
       if pk = pk' then
         mgu_aux ul l
       else
         raise NoUnifier
    | ((VVerKey (_, vk), VVerKey (_, vk')) :: l) ->
       if vk = vk' then
         mgu_aux ul l
       else
         raise NoUnifier
    | ((VPair (_, u, u'), VPair (_, v, v')) :: l) ->
         mgu_aux ul ((u,v) :: (u',v') :: l)
    | ((VInl (_, u), VInl (_, v)) :: l) ->
         mgu_aux ul ((u,v) :: l)
    | ((VInr (_, u), VInr (_, v)) :: l) ->
         mgu_aux ul ((u,v) :: l)
    | ((VHash (_, u), VHash (_, v)) :: l) ->
         mgu_aux ul ((u,v) :: l)
    | ((VSEnc (_, u, k), VSEnc (_, v, k')) :: l) ->
       if k = k' then
         mgu_aux ul ((u,v) :: l)
       else
         raise NoUnifier
    | ((VAEnc (_, u, k), VAEnc (_, v, k')) :: l) ->
       if k = k' then
         mgu_aux ul ((u,v) :: l)
       else
         raise NoUnifier
    | ((VSign (_, u, k), VSign (_, v, k')) :: l) ->
       if k = k' then
         mgu_aux ul ((u,v) :: l)
       else
         raise NoUnifier
    (* incompatible head symbols *)
    | _ -> raise NoUnifier
  in
  mgu_aux [] ul

let unify (u:term) (v:term) : substitution option =
  try
    let mu = mgu [u,v] in
    Some (List.map
            (fun (u,v) -> match u with
                          | VVar (_, x) -> (x,v)
                          | _ -> failwith "Illegal state after unification: substitution contain other things than variables")
            mu)
  with
    NoUnifier -> None

                 
               
let check_mgu (mu:substitution) (g:var_env) (sd:side) : substitution option =
  let mumu = List.map
               (fun (x,u) ->
                 match lookup_var g x with
                 | Some (TRefine (VNonce (e1,NonceStar n1), VNonce (e2, NonceStar n2))) ->
                    (match u with
                     | VVar _ -> (true, None)
                     | VNonce (_, Nonce n) ->
                        let n' = if sd = Left then n1 else n2 in
                        let (e'', n'') = if sd = Left then (e2,n2) else (e1,n1) in
                        if n = expand_nonce n' 1
                        then (true, Some (x,VNonce (e'', Nonce (expand_nonce n'' 1))))
                        else
                          (if n = expand_nonce n' 2
                           then (true, Some (x, VNonce (e'', Nonce (expand_nonce n'' 2))))
                           else (false, None))
                     | _ -> (false, None))
                 | Some (TLabel LL) ->
                    (match u with
                     | VNonce (_, n) ->
                        (true, Some (x,u))
                     | _ -> (true, None))
                 |_ -> (true, None))
               mu in
  let b = List.for_all fst mumu in
  if b then
    Some (List.map (fun (_,uu) -> match uu with
                                  | Some vv -> vv
                                  | None -> failwith "cannot happen (check_mgu)")
                   (List.filter (fun (_,uu)-> match uu with
                                              | Some _ -> true
                                              | None -> false)
                                mumu))
  else None
                



let print_substitution (s:substitution) : unit =
  let l = List.map (fun (x,u) -> (var2string x) ^ "->" ^ (term2string u)) s in
  let s = String.concat "; " l in
  Printf.printf "[[%s]]\n" s
  
let step4' (g:var_env) (c:constr) (c':constr) : unit = 
(*   print_verbose 2 ("Checking " ^ constr2string c ^ " <-> " ^ constr2string c'); *)
  let u,v = c in 
  let u',v' = c' in
  (match unify u u' with
   | Some mu ->
      (match check_mgu mu g Left with
       | Some theta ->
          let v1 = substitute theta v in
          let v1' = substitute theta v' in
          if not (same_term v1 v1') then            
            raise (Constraint_Error (constr2string c ^ " <> " ^ constr2string c' ^ " ( " ^ string_of_bool true ^ ", " ^ string_of_bool false ^ ")"))
       | _ -> ())
   | _ -> ());
    (match unify v v' with
     | Some mu ->
        (match check_mgu mu g Right with
         | Some theta ->
            let u1 = substitute theta u in
            let u1' = substitute theta u' in
            if not (same_term u1 u1') then
              raise (Constraint_Error (constr2string c ^ " <> " ^ constr2string c' ^ " ( " ^ string_of_bool false ^ ", " ^ string_of_bool true ^ ")"))
         | _ -> ())
     | _ -> ())

(* Step 4 from algorithm *)
let rec step4 (e:nonce_env) (e':nonce_env) (g:var_env) (d:key_env) (cs:constr_set) : bool =  
  match cs with 
  | c :: cs -> 
    step3 e e' g d c && 
      (List.iter (fun c' -> step4' g c c') cs;
       step4 e e' g d cs)
  | [] -> true


let rec check_consistency_cs (e:nonce_env) (e':nonce_env) (g:var_env) (d:key_env) (cs: constr_set) : bool = 
  (*  print_verbose 2 ("Checking consistency of : \n " ^ constrset2string cs);*)
  if List.for_all (step0 e e g d) cs then
    let cs' = List.map (step1 g) cs in 
    let cs'' = List.flatten (List.map (fun c -> step2 d [c] []) cs') in
    step4 e e' g d cs''
  else
    raise (Constraint_Error "Step 0 failed")

let rec check_consistency (e1:nonce_env) (e2:nonce_env) (d:key_env) (css : constr_set_set) (css' : constr_set_set) : bool = 
  print_verbose 0 "Replicating constraints...";
  let css1 = expand_repl_constr_set_set css' 1 in 
  let css2 = expand_repl_constr_set_set css' 2 in 
  let css3 = expand_repl_constr_set_set css 1 in 
  let e1' = expand_repl_nonce_env e1 in 
  let e2' = expand_repl_nonce_env e2 in 
  print_verbose 0 "Checking consistency...";
  List.for_all (fun (cs1,g1) -> 
    List.for_all (fun (cs2,g2) -> 
      List.fold_left (fun b (cs3,g3) -> 
        (match var_env_union g2 g3 with 
        | Some g' ->
          (match var_env_union g1 g' with
          | Some g'' ->
            b && check_consistency_cs e1' e2' g'' d (cs1 @ cs2 @ cs3) 
          | None -> b) 
        | None -> b))
      true css3) 
    css2) 
  css1


